import xarray as xr
import numpy as np

from .filters import SpymFilters
from .level import SpymLevel
from .plotting import SpymPlotting

@xr.register_dataarray_accessor("spym")
class _SpymDataArray:
    '''spym class extending xarray DataArray

    '''

    def __init__(self, xarray_dr):
        self._dr = xarray_dr

        self._bkg = np.zeros(self._dr.data.shape)
        self._mask = np.zeros(self._dr.data.shape, dtype=bool)

        # Initialize spym classes
        self.level = SpymLevel(self)
        self.filters = SpymFilters(self)
        self.plotting = SpymPlotting(self)

        # Expose some useful methods
        self.fixzero = self.level.fixzero
        self.plane = self.level.plane
        self.align = self.level.align
        self.destripe = self.filters.destripe
        self.plot = self.plotting.plot

    @property
    def background(self):
        ''' Background.

        '''
        return self._bkg

    @property
    def mask(self):
        ''' Mask.

        '''
        return self._mask

@xr.register_dataset_accessor("spym")
class _SpymDataset:
    '''spym class extending xarray Dataset

    '''

    def __init__(self, xarray_ds):
        self._ds = xarray_ds



from spym.process import filters as spym_filters

class SpymFilters():
    ''' Filters.

    '''

    def __init__(self, spym_instance):
        self._spym = spym_instance

    def gauss(self, **kwargs):
        ''' Apply Gaussian smoothing filter.

        Args:
            size: size of the filter in pixels.

        '''

        self._spym._dr.data = spym_filters.gauss(self._spym._dr.data, **kwargs)

    def median(self, **kwargs):
        ''' Apply median smoothing filter.

        Args:
            size: size of the filter in pixels.

        '''

        self._spym._dr.data = spym_filters.median(self._spym._dr.data, **kwargs)

    def mean(self, **kwargs):
        ''' Apply mean smoothing filter.

        Args:
            size: size of the filter in pixels.

        '''

        self._spym._dr.data = spym_filters.mean(self._spym._dr.data, **kwargs)

    def sharpen(self, **kwargs):
        ''' Apply a sharpening filter.

        Args:
            size: size of the filter in pixels.
            alpha: weight.

        '''

        self._spym._dr.data = spym_filters.sharpen(self._spym._dr.data, **kwargs)

    def destripe(self, **kwargs):
        ''' Find and remove scan stripes by averaging neighbourhood lines.

        Args:
            min_length: only scars that are as long or longer than this value (in pixels) will be marked.
            hard_threshold: the minimum difference of the value from the neighbouring upper and lower lines to be considered a defect.
            soft_threshold: values differing at least this much do not form defects themselves, but they are attached to defects obtained from the hard threshold if they touch one.
            sign: whether mark stripes with positive values, negative values or both.
            rel_threshold: the minimum difference of the value from the neighbouring upper and lower lines to be considered a defect (in physical values). Overwrite hard_threshold.

        Returns:
            destriped 2d array.

        '''

        if not self._spym._dr.data.ndim == 2:
            print("The DataArray is not an image. Abort.")
            return

        self._spym._dr.data, self._spym._mask = spym_filters.destripe(self._spym._dr.data.astype(float), **kwargs)



import os

def load(filename, scaling=True):
    ''' Import data from common SPM file formats.

    Currently supported file formats are:
        * NeXus (.nx, .nxs). Package nxarray is needed.
        * RHK (.sm4). Package rhksm4 is needed.
        * Omicron Scala (.par). Package omicronscala is needed.

    Args:
        filename: path to the SPM file.
        scaling: if True convert data to physical units (default), if False keep raw data.

    Returns:
        xarray Dataset with data and metadata.

    '''

    if filename.endswith(".nx") or filename.endswith(".nxs"):
        try:
            import nxarray
            ds = nxarray.load(filename)
        except:
            print("Error: nxarray package is needed to open .nx/.nxs files.")
            ds = None

    if filename.endswith(".par"):
        try:
            import omicronscala
            ds = omicronscala.to_dataset(filename, scaling=scaling)
        except:
            print("Error: omicronscala package is needed to open .par files.")
            ds = None

    if filename.endswith(".sm4"):
        try:
            import rhksm4
            ds = rhksm4.to_dataset(filename, scaling=scaling)
        except:
            print("Error: rhksm4 package is needed to open .sm4 files.")
            ds = None

    if ds is not None:
        for dr in ds:
            ds[dr].attrs["filename"] = os.path.basename(filename)

    return ds



import matplotlib.pyplot as plt

class SpymPlotting():
    ''' Plotting.
    
    '''

    def __init__(self, spym_instance):
        self._spym = spym_instance

    def plot(self, title=None, **kwargs):
        ''' Plot data with custom parameters.

        Args:
            title: title of the figure (string). By default gives some basic information on the data plotted. Pass an empty string to disable it.
            **kwargs: any argument accepted by xarray.plot() function.

        '''

        dr = self._spym._dr

        # Create figure
        # xarray plot() wraps:
        #   - matplotlib.pyplot.plot() for 1d arrays
        #   - matplotlib.pyplot.pcolormesh() for 2d arrays
        #   - matplotlib.pyplot.hist() for anything else
        plot = dr.plot()

        # Set figure title
        if title is None:
            title = self.format_title()
        plt.title(title)

        # Set images properties (2d arrays)
        # plot is an instance of matplotlib.collections.QuadMesh
        if dr.data.ndim == 2 and not isinstance(plot, list):
            fig = plot.get_figure()
            ax = plot.axes
            # Fit figure pixel size to image
            fig_width, fig_height = self.fit_figure_to_image(fig, dr.data, ax)
            fig.set_size_inches(fig_width, fig_height)

            # Apply colormap
            plot.set_cmap('afmhot')

        plt.draw()

    def format_title(self):
        ''' Provide a title from the metadata of the DataArray.

        '''

        title = ""
        attrs = self._spym._dr.attrs

        if "filename" in attrs:
            title += attrs["filename"] + "\n"

        title += "{:.2f} {}, {:.2f} {}".format(
            attrs["bias"],
            attrs["bias_units"],
            attrs["setpoint"],
            attrs["setpoint_units"])

        return title

    def fit_figure_to_image(self, figure, image, axis=None):
        ''' Calculate figure size so that plot (matplotlib axis) pixel size is equal to the image size.

        Args:
            figure: matplotlib Figure instance.
            image: 2d numpy array.
            axis: axis of the figure to adapt, if None takes the first (or only) axis.

        Returns:
            adapted width and height of the figure in inches.

        '''

        if axis is None:
            axis = figure.axes[0]
        bounds = axis.bbox.bounds

        im_width, im_height = image.shape

        width_scale = im_width/bounds[2]
        height_scale = im_height/bounds[3]

        fig_width, fig_height = figure.get_size_inches()

        return fig_width*width_scale, fig_height*height_scale



from spym.process import level as spym_level

class SpymLevel():
    ''' Level.

    '''

    def __init__(self, spym_instance):
        self._spym = spym_instance

    def fixzero(self, **kwargs):
        ''' Add a constant to all the data to move the minimum (or the mean value) to zero.

        Args:
            to_mean: bool, optional. If true move mean value to zero, if false move mimimum to zero (default).

        '''

        self._spym._dr.data = spym_level.fixzero(self._spym._dr.data, **kwargs)

    def plane(self, **kwargs):
        '''Corrects for sample tilting by subtraction of a plane.

        '''

        if not self._spym._dr.data.ndim == 2:
            print("The DataArray is not an image. Abort.")
            return

        self._spym._dr.data, self._spym._bkg = spym_level.plane(self._spym._dr.data.astype(float), **kwargs)

    def align(self, **kwargs):
        '''Align rows.

        Args:
            baseline: defines how baselines are estimated; 'mean' (default), 'median', 'poly'.
            axis: axis along wich calculate the baselines.
            poly_degree: polnomial degree if baseline='poly'.

        '''

        if not self._spym._dr.data.ndim == 2:
            print("The DataArray is not an image. Abort.")
            return

        self._spym._dr.data, self._spym._bkg = spym_level.align(self._spym._dr.data.astype(float), **kwargs)
